/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ArticulatedObjectStateProviderExample.h"

#include <IceUtil/Time.h>
#include <RobotAPI/interface/core/articulated_object/provider.h>
#include <SimoxUtility/algorithm/string/string_tools.h>

#include <ArmarXCore/core/time/CycleUtil.h>

#include <RobotAPI/libraries/ArmarXObjects/ice_conversions.h>
#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/core/Pose.h>

#include <RobotAPI/interface/core/articulated_object/types.h>
#include <cmath>
#include <vector>

namespace armarx::articulated_object
{

    armarx::PropertyDefinitionsPtr ArticulatedObjectStateProviderExample::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs =
            new ComponentPropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        // defs->optional(initialObjectIDs, "Objects", "Object IDs of objects to be tracked.")
        //         .map(simox::alg::join(initialObjectIDs, ", "), initialObjectIDs);

        return defs;
    }

    std::string ArticulatedObjectStateProviderExample::getDefaultName() const
    {
        return "ArticulatedObjectStateProviderExample";
    }

    void ArticulatedObjectStateProviderExample::onInitComponent()
    {
        providerInfo.supportedObjects.push_back(dishwasherId);
    }

    void ArticulatedObjectStateProviderExample::onConnectComponent()
    {
        poseEstimationTask = new SimpleRunningTask<>([this]()
        {
            this->poseEstimationTaskRun();
        });
        poseEstimationTask->start();
    }

    void ArticulatedObjectStateProviderExample::onDisconnectComponent() {}

    void ArticulatedObjectStateProviderExample::onExitComponent() {}

    provider::ProviderInfo ArticulatedObjectStateProviderExample::getProviderInfo(const Ice::Current&)
    {
        return providerInfo;
    }

    std::vector<std::string> objectNames(const data::ArticulatedObjectIDs& ids) noexcept
    {

        const auto objectName = [](const data::ArticulatedObjectID & id)
        {
            return id.name;
        };


        std::vector<std::string> names;
        names.reserve(ids.size());

        std::transform(ids.begin(), ids.end(), std::back_inserter(names), objectName);

        return names;
    }

    provider::TrackingRequestResponse ArticulatedObjectStateProviderExample::trackObjects(const provider::TrackingRequest& input, const Ice::Current&)
    {
        ARMARX_INFO << "Requested object IDs for " << input.relativeTimeoutMS
                    << " ms: " << objectNames(input.objectIDs);
        {
            std::scoped_lock lock(requestedObjectsMutex);
            requestedObjects = input.objectIDs;
        }

        provider::TrackingRequestResponse response;

        // All requests are successful.
        for (const auto& id : input.objectIDs)
        {
            response.status[id] = (id == dishwasherId);
        }

        return response;
    }

    void ArticulatedObjectStateProviderExample::poseEstimationTaskRun()
    {
        CycleUtil cycle(IceUtil::Time::milliSeconds(100));
        IceUtil::Time start = TimeUtil::GetTime();

        while (poseEstimationTask && !poseEstimationTask->isStopped())
        {

            ARMARX_INFO << "Reporting articulated objects";


            IceUtil::Time now = TimeUtil::GetTime();
            const float t     = float((now - start).toSecondsDouble());

            // move joints at certain frequency
            const float k = (1 + std::sin(t / (M_2_PIf32 * 0.2))) / 2; // in [0,1]

            const articulated_object::data::ArticulatedObject articulatedObject
            {
                .id = dishwasherId,
                .config = {
                    .timestamp  = TimeUtil::GetTime().toSecondsDouble(),
                    .globalPose = new FramedPose(),
                    .jointMap   = {{"dishwasher_door_joint", 0.2 * k},
                        {"drawer_joint", 350 * k}
                    }
                }};

            ARMARX_VERBOSE << "Reporting articulated object";
            articulatedObjectTopic->reportArticulatedObjects(getName(), {articulatedObject});

            cycle.waitForCycleDuration();
        }
    }
} // namespace armarx::articulated_object
