/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ObjectMemory
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ObjectMemory.h"

#include <RobotAPI/interface/core/articulated_object/topic.h>
#include <mutex>


namespace armarx::armem::server::obj
{

    const std::string ObjectMemory::defaultMemoryName = "Object";


    armarx::PropertyDefinitionsPtr ObjectMemory::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs(new ComponentPropertyDefinitions(getConfigIdentifier()));

        // Offer
        defs->topic(debugObserver);

        // Subscribe
        defs->topic<objpose::ObjectPoseTopic>(); // "ObjectPoseTopic", "ObjectPoseTopicName", "Name of the Object Pose Topic.");

        // Use
        // defs->component(kinematicUnitObserver);  // Optional dependency.
        defs->defineOptionalProperty<std::string>("cmp.KinematicUnitObserverName", "KinematicUnitObserver",
                "Name of the kinematic unit observer.");

        const std::string prefix = "mem.";

        workingMemory.name() = defaultMemoryName;
        defs->optional(workingMemory.name(), prefix + "MemoryName", "Name of this memory server.");

        classSegment.defineProperties(defs, prefix + "cls.");
        instance::SegmentAdapter::defineProperties(defs, prefix + "inst.");

        return defs;
    }

    ObjectMemory::ObjectMemory() :
        server::ComponentPluginUser(),
        instance::SegmentAdapter(server::ComponentPluginUser::iceMemory,
                                 server::ComponentPluginUser::workingMemoryMutex),
        classSegment(server::ComponentPluginUser::iceMemory,
                     server::ComponentPluginUser::workingMemoryMutex)
    {
    }

    std::string ObjectMemory::getDefaultName() const
    {
        return "ObjectMemory";
    }

    void ObjectMemory::onInitComponent()
    {
        workingMemory.name() = defaultMemoryName;

        instance::SegmentAdapter::init();

        try
        {
            classSegment.init();
        }
        catch (const LocalException& e)
        {
            ARMARX_ERROR << "Failed to init class segment. Reason: \n" << e.what();
        }
        catch (const std::exception& e)
        {
            ARMARX_ERROR << "Failed to init class segment. Reason: \n" << e.what();
        }
        catch (...)
        {
            ARMARX_ERROR << "Failed to init class segment for unknown reason.";
        }
    }

    void ObjectMemory::onConnectComponent()
    {
        // onConnect can be called multiple times, but addRobot will fail if called more than once with the same ID
        // So we need to always make sure to guard a call to addRobot
        const std::string robotKey = "robot";
        VirtualRobot::RobotPtr robot = RobotState::hasRobot(robotKey)
                                       ? RobotState::getRobot(robotKey)
                                       : RobotState::addRobot(robotKey, VirtualRobot::RobotIO::RobotDescription::eStructure);

        robotStateComponent = RobotState::getRobotStateComponent();

        getProxyFromProperty(kinematicUnitObserver, "cmp.KinematicUnitObserverName", false, "", false);

        instance::SegmentAdapter::connect(
            robotStateComponent,
            robot,
            kinematicUnitObserver,
            ArVizComponentPluginUser::arviz,
            debugObserver
        );
        classSegment.connect(
            ArVizComponentPluginUser::arviz
        );


        {
            std::lock_guard g(server::ComponentPluginUser::workingMemoryMutex);
            server::ComponentPluginUser::workingMemory.addCoreSegment("ArticulatedObjectInstance"); // TODO , arondto::::toInitialAronType());
            server::ComponentPluginUser::workingMemory.addCoreSegment("ArticulatedObjectClass"); // TODO , arondto::::toInitialAronType());
        }

        createRemoteGuiTab();
        RemoteGui_startRunningTask();

    }

    void ObjectMemory::onDisconnectComponent()
    {
    }

    void ObjectMemory::onExitComponent()
    {
    }


    void ObjectMemory::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        tab.instance.setup(*this);
        tab.clazz.setup(classSegment);

        HBoxLayout segments =
        {
            tab.instance.group,
            tab.clazz.group
        };
        VBoxLayout root =
        {
            segments,
            VSpacer()
        };
        RemoteGui_createTab(Component::getName(), root, &tab);
    }

    void ObjectMemory::RemoteGui_update()
    {
        // Non-atomic variables need to be guarded by a mutex if accessed by multiple threads
        tab.instance.update(*this);
        tab.clazz.update(classSegment);
    }

}

