#include "Reader.h"

#include <ArmarXCore/core/logging/Logging.h>

#include "query/Builder.h"
#include "query/query_fns.h"


namespace armarx::armem::client
{

    Reader::Reader(server::ReadingMemoryInterfacePrx memory) : memoryPrx(memory)
    {
    }


    QueryResult
    Reader::query(const QueryInput& input) const
    {
        return QueryResult::fromIce(query(input.toIce()));
    }

    armem::query::data::Result
    Reader::query(const armem::query::data::Input& input) const
    {
        auto makeErrorMsg = [](auto e)
        {
            std::stringstream ss;
            ss << "Memory query failed.\nReason: " << e.what();
            return ss.str();
        };

        armem::query::data::Result result;
        try
        {
            result = memoryPrx->query(input);
        }
        catch (const Ice::ConnectionRefusedException& e)
        {
            result.errorMessage = makeErrorMsg(e);
        }
        catch (const Ice::ConnectionLostException& e)
        {
            result.errorMessage = makeErrorMsg(e);
        }
        catch (const Ice::NotRegisteredException& e)
        {
            result.errorMessage = makeErrorMsg(e);
        }
        return result;
    }

    QueryResult Reader::query(armem::query::data::MemoryQueryPtr query, DataMode dataMode) const
    {
        return this->query(armem::query::data::MemoryQuerySeq{query}, dataMode);
    }

    QueryResult Reader::query(const armem::query::data::MemoryQuerySeq& queries, DataMode dataMode) const
    {
        QueryInput input;
        input.memoryQueries = queries;
        input.dataMode = dataMode;
        return this->query(input);
    }

    QueryResult Reader::getAll(DataMode dataMode) const
    {
        using namespace client::query_fns;

        query::Builder qb(dataMode);
        qb.coreSegments(all()).providerSegments(all()).entities(all()).snapshots(all());

        return this->query(qb.buildQueryInput());
    }

    QueryResult Reader::getLatestSnapshots(DataMode dataMode) const
    {
        using namespace client::query_fns;

        query::Builder qb(dataMode);
        qb.coreSegments(all()).providerSegments(all()).entities(all()).snapshots(latest());

        return this->query(qb.buildQueryInput());
    }

    void
    Reader::updated(const std::vector<MemoryID>& updatedSnapshotIDs) const
    {
        for (const auto& [subscription, callback] : callbacks)
        {
            std::vector<MemoryID> matchingSnapshotIDs;

            for (const MemoryID& updatedSnapshotID : updatedSnapshotIDs)
            {
                if (contains(subscription, updatedSnapshotID))
                {
                    matchingSnapshotIDs.push_back(updatedSnapshotID);
                }
            }

            if (not matchingSnapshotIDs.empty())
            {
                callback(subscription, matchingSnapshotIDs);
            }
        }
    }

    data::StoreResult
    Reader::readAndStore(const data::StoreInput& input) const
    {
        server::StoringMemoryInterfacePrx storingMemoryPrx = server::StoringMemoryInterfacePrx::checkedCast(memoryPrx);
        if (storingMemoryPrx)
        {
            return storingMemoryPrx->store(input);
        }
        else
        {
            ARMARX_WARNING << "Could not store a query into the LTM. It seems like the Memory does not implement the StoringMemoryInterface.";
            return {};
        }
    }


    void
    Reader::subscribe(const MemoryID& id, callback callback)
    {
        callbacks[id] = callback;
    }


    void
    Reader::setReadingMemory(server::ReadingMemoryInterfacePrx memory)
    {
        this->memoryPrx = memory;
    }

}
