/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::armem
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::armem

#define ARMARX_BOOST_TEST

#include <RobotAPI/Test.h>
#include "../server/query_proc/workingmemory/EntityQueryProcessor.h"
#include "../error.h"


#include <iostream>

#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <SimoxUtility/algorithm/string.h>


namespace armem = armarx::armem;
namespace aron = armarx::aron;
namespace query = armarx::armem::query::data;
using EntityQueryProcessor = armarx::armem::wm::query_proc::EntityQueryProcessor;


namespace ArMemQueryTest
{

    class UnknownEntityQuery : public query::EntityQuery
    {
    };

    struct Fixture
    {
        armem::wm::Entity entity;

        armem::wm::query_proc::EntityQueryProcessor processor;
        query::EntityQueryPtr entityQuery;

        std::vector<armem::wm::Entity> results;


        Fixture()
        {
            entity = armem::wm::Entity("entity");

            armem::wm::EntitySnapshot snapshot;
            snapshot.time() = armem::Time::microSeconds(1000);
            entity.addSnapshot(snapshot);
            snapshot.time() = armem::Time::microSeconds(2000);
            entity.addSnapshot(snapshot);
            snapshot.time() = armem::Time::microSeconds(3000);
            entity.addSnapshot(snapshot);
            snapshot.time() = armem::Time::microSeconds(4000);
            entity.addSnapshot(snapshot);
            snapshot.time() = armem::Time::microSeconds(5000);
            entity.addSnapshot(snapshot);
        }
        ~Fixture()
        {
        }


        template <class QueryT>
        void addResults(QueryT query = {})
        {
            entityQuery = new QueryT(query);

            results.emplace_back(processor.process(query, entity));
            results.emplace_back(processor.process(*entityQuery, entity));
        }
    };

}


BOOST_FIXTURE_TEST_SUITE(ArMemQueryTest, ArMemQueryTest::Fixture)



BOOST_AUTO_TEST_CASE(test_entity_UnknownQueryType)
{
    BOOST_CHECK_THROW(processor.process(UnknownEntityQuery(), entity), armem::error::UnknownQueryType);
    std::string msg;
    try
    {
        processor.process(UnknownEntityQuery(), entity);
    }
    catch (const armem::error::UnknownQueryType& e)
    {
        msg = e.what();
    }
    BOOST_TEST_MESSAGE(msg);
    BOOST_CHECK(simox::alg::contains(msg, "UnknownEntityQuery"));
}



BOOST_AUTO_TEST_CASE(test_entity_Single_latest)
{
    addResults(query::entity::Single());
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());
        BOOST_CHECK_EQUAL(result.size(), 1);
        BOOST_CHECK_EQUAL(result.history().begin()->second.time(), entity.getLatestSnapshot().time());
        BOOST_CHECK_NE(&result.history().begin()->second, &entity.history().rbegin()->second);
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Single_existing)
{
    addResults(query::entity::Single { 3000 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());
        BOOST_CHECK_EQUAL(result.size(), 1);
        BOOST_CHECK_EQUAL(result.history().begin()->second.time(), armem::Time::microSeconds(3000));
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Single_non_existing)
{
    addResults(query::entity::Single { 3500 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());
        BOOST_CHECK_EQUAL(result.size(), 0);
    }
}


BOOST_AUTO_TEST_CASE(test_entity_All)
{
    addResults(query::entity::All());
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());
        BOOST_CHECK_EQUAL(result.size(), entity.size());
        auto jt = entity.history().begin();
        for (auto it = result.history().begin(); it != result.history().end(); ++it, ++jt)
        {
            BOOST_CHECK_EQUAL(it->first, jt->first);
            BOOST_CHECK_EQUAL(it->second.time(), jt->second.time());
            BOOST_CHECK_NE(&it->second, &jt->second);
        }
    }
}



BOOST_AUTO_TEST_CASE(test_entity_TimeRange_slice)
{
    addResults(query::entity::TimeRange{ 1500, 3500 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), 2);

        std::vector<armem::Time> times = simox::alg::get_keys(result.history());
        std::vector<armem::Time> expected
        {
            armem::Time::microSeconds(2000), armem::Time::microSeconds(3000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_TimeRange_exact)
{
    addResults(query::entity::TimeRange{ 2000, 4000 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), 3);

        std::vector<armem::Time> times = simox::alg::get_keys(result.history());
        std::vector<armem::Time> expected
        {
            armem::Time::microSeconds(2000), armem::Time::microSeconds(3000), armem::Time::microSeconds(4000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}

BOOST_AUTO_TEST_CASE(test_entity_TimeRange_all)
{
    addResults(query::entity::TimeRange{ 0, 10000 });
    addResults(query::entity::TimeRange{ -1, -1 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), entity.size());

        std::vector<armem::Time> times = simox::alg::get_keys(result.history());
        std::vector<armem::Time> expected = simox::alg::get_keys(entity.history());
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_TimeRange_empty)
{
    addResults(query::entity::TimeRange{ 2400, 2600 });
    addResults(query::entity::TimeRange{ 6000, 1000 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), 0);
    }
}


BOOST_AUTO_TEST_CASE(test_entity_TimeRange_from_start)
{
    addResults(query::entity::TimeRange{ -1, 2500 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), 2);

        std::vector<armem::Time> times = simox::alg::get_keys(result.history());
        std::vector<armem::Time> expected
        {
            armem::Time::microSeconds(1000), armem::Time::microSeconds(2000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_TimeRange_to_end)
{
    addResults(query::entity::TimeRange{ 2500, -1 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), 3);

        std::vector<armem::Time> times = simox::alg::get_keys(result.history());
        std::vector<armem::Time> expected
        {
            armem::Time::microSeconds(3000), armem::Time::microSeconds(4000), armem::Time::microSeconds(5000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_negative_index_semantics)
{
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(0, 0), 0);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(0, 1), 0);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(0, 10), 0);

    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(1, 1), 0);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(5, 1), 0);

    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(1, 5), 1);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(5, 5), 4);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(5, 6), 5);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(5, 10), 5);

    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-1, 0), 0);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-5, 0), 0);

    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-1, 1), 0);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-5, 1), 0);

    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-1, 2), 1);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-5, 2), 0);

    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-1, 5), 4);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-5, 5), 0);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-5, 6), 1);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-5, 10), 5);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-10, 10), 0);
    BOOST_CHECK_EQUAL(EntityQueryProcessor::negativeIndexSemantics(-20, 10), 0);
}


BOOST_AUTO_TEST_CASE(test_entity_IndexRange_all_default)
{
    addResults(query::entity::IndexRange());
    addResults(query::entity::IndexRange(0, -1));
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), entity.size());

        std::vector<armem::Time> times = simox::alg::get_keys(result.history());
        std::vector<armem::Time> expected = simox::alg::get_keys(entity.history());
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }

}


BOOST_AUTO_TEST_CASE(test_entity_IndexRange_slice)
{
    BOOST_REQUIRE_EQUAL(entity.size(), 5);
    addResults(query::entity::IndexRange{  1,  3 }); // => [1, 2, 3]
    addResults(query::entity::IndexRange{  1, -2 }); // 5 - 2 = 3
    addResults(query::entity::IndexRange{ -4,  3 }); // 5 - 4 = 1
    addResults(query::entity::IndexRange{ -4, -2 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), 3);

        std::vector<armem::Time> times = simox::alg::get_keys(result.history());
        std::vector<armem::Time> expected
        {
            armem::Time::microSeconds(2000), armem::Time::microSeconds(3000), armem::Time::microSeconds(4000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_IndexRange_empty_range)
{
    BOOST_REQUIRE_EQUAL(entity.size(), 5);
    addResults(query::entity::IndexRange{  1,  0 });
    addResults(query::entity::IndexRange{  2,  1 });
    addResults(query::entity::IndexRange{  5,  3 });
    addResults(query::entity::IndexRange{  4, -3 });  // 5-3 = 2
    addResults(query::entity::IndexRange{  3, -3 });
    addResults(query::entity::IndexRange{  1, -5 });

    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());

        BOOST_CHECK_EQUAL(result.size(), 0);
    }
}


BOOST_AUTO_TEST_CASE(test_entity_IndexRange_empty_entity)
{
    entity.clear();
    BOOST_REQUIRE_EQUAL(entity.size(), 0);
    addResults(query::entity::IndexRange{  0,  0 });
    addResults(query::entity::IndexRange{  0, 10 });
    addResults(query::entity::IndexRange{-10, -1 });
    addResults(query::entity::IndexRange{  2,  5 });
    addResults(query::entity::IndexRange{  3, -3 });
    addResults(query::entity::IndexRange{ -1, 10 });

    BOOST_REQUIRE_GT(results.size(), 0);

    for (const armem::wm::Entity& result : results)
    {
        BOOST_CHECK_EQUAL(result.name(), entity.name());
        BOOST_CHECK_EQUAL(result.size(), 0);
    }
}


BOOST_AUTO_TEST_SUITE_END()
