#include "InstanceView.h"

#include <QAction>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QImage>
#include <QLabel>
#include <QLayout>
#include <QMenu>
#include <QSplitter>
#include <QTreeWidget>
#include <QVBoxLayout>

#include <SimoxUtility/algorithm/string.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/aron/core/navigator/data/complex/NDArray.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Object.h>

#include <RobotAPI/libraries/armem_gui/gui_utils.h>
#include <RobotAPI/libraries/armem_gui/instance/ImageView.h>
#include <RobotAPI/libraries/armem_gui/instance/serialize_path.h>
#include <RobotAPI/libraries/armem_gui/instance/tree_builders/DataTreeBuilder.h>
#include <RobotAPI/libraries/armem_gui/instance/tree_builders/TypedDataTreeBuilder.h>


namespace armarx::armem::gui::instance
{

    InstanceView::InstanceView()
    {
        QLayout* layout = new QVBoxLayout();
        this->setLayout(layout);
        int margin = 3;
        layout->setContentsMargins(margin, margin, margin, margin);

        splitter = new QSplitter(Qt::Orientation::Vertical);
        layout->addWidget(splitter);

        tree = new QTreeWidget(this);
        splitter->addWidget(tree);

        QStringList columns;
        columns.insert(int(Columns::KEY), "Key");
        columns.insert(int(Columns::VALUE), "Value");
        columns.insert(int(Columns::TYPE), "Type");
        tree->setColumnCount(columns.size());
        tree->setHeaderLabels(columns);

        tree->header()->resizeSection(int(Columns::KEY), 250);
        tree->header()->resizeSection(int(Columns::VALUE), 150);

        treeItemInstanceID = new QTreeWidgetItem({"Instance ID"});
        treeItemInstanceID->addChild(new QTreeWidgetItem({"Memory Name"}));
        treeItemInstanceID->addChild(new QTreeWidgetItem({"Core Segment Name"}));
        treeItemInstanceID->addChild(new QTreeWidgetItem({"Provider Segment Name"}));
        treeItemInstanceID->addChild(new QTreeWidgetItem({"Entity Name"}));
        treeItemInstanceID->addChild(new QTreeWidgetItem({"Timestamp"}));
        treeItemInstanceID->addChild(new QTreeWidgetItem({"Instance Index"}));

        treeItemMetadata = new QTreeWidgetItem({"Metadata"});
        treeItemMetadata->addChild(new QTreeWidgetItem({"Confidence"}));
        treeItemMetadata->addChild(new QTreeWidgetItem({"Time Created"}));
        treeItemMetadata->addChild(new QTreeWidgetItem({"Time Sent"}));
        treeItemMetadata->addChild(new QTreeWidgetItem({"Time Arrived"}));

        treeItemData = new QTreeWidgetItem({"Data"});

        QList<QTreeWidgetItem*> items = {treeItemInstanceID, treeItemMetadata, treeItemData};
        tree->addTopLevelItems(items);
        for (auto* item : items)
        {
            item->setExpanded(true);
        }

        tree->setContextMenuPolicy(Qt::CustomContextMenu);
        connect(tree, &QTreeWidget::customContextMenuRequested, this, &This::prepareTreeContextMenu);
    }

    void InstanceView::setStatusLabel(QLabel* statusLabel)
    {
        this->statusLabel = statusLabel;
    }

    void InstanceView::setUseTypeInfo(bool enable)
    {
        this->useTypeInfo = enable;
        update();
    }


    void InstanceView::update(const MemoryID& id, const wm::Memory& memory)
    {
        aron::typenavigator::ObjectNavigatorPtr aronType = nullptr;
        const armem::wm::EntityInstance* instance = nullptr;
        try
        {
            instance = &memory.getEntityInstance(id);
            if (useTypeInfo)
            {
                aronType = memory.getCoreSegment(id.coreSegmentName).getProviderSegment(id.providerSegmentName).aronType();
            }
        }
        catch (const armem::error::ArMemError& e)
        {
            showErrorMessage(e.what());
            return;
        };
        if (instance)
        {
            update(*instance, aronType);
        }
    }


    void InstanceView::update(const wm::EntityInstance& instance, aron::typenavigator::ObjectNavigatorPtr aronType)
    {
        currentInstance = instance;
        currentAronType = aronType;
        update();
    }

    void InstanceView::update()
    {
        if (currentInstance)
        {
            updateInstanceID(currentInstance->id());
            updateMetaData(currentInstance->metadata());
            updateData(currentInstance->data(), currentAronType);
            updateImageView(currentInstance->data());

            emit updated();
        }
    }


    void InstanceView::updateInstanceID(const MemoryID& id)
    {
        const std::vector<std::string> items = id.getAllItems();
        ARMARX_CHECK_EQUAL(treeItemInstanceID->childCount(), static_cast<int>(items.size()));
        int i = 0;
        for (const std::string& item : items)
        {
            treeItemInstanceID->child(i++)->setText(int(Columns::VALUE), QString::fromStdString(item));
        }
    }

    void InstanceView::updateData(const aron::datanavigator::DictNavigatorPtr& data, aron::typenavigator::ObjectNavigatorPtr aronType)
    {
        if (!data)
        {
            armarx::gui::clearItem(treeItemData);
            QTreeWidgetItem* item = new QTreeWidgetItem({"(No data.)"});
            treeItemData->addChild(item);
        }
        else if (useTypeInfo && aronType)
        {
            TypedDataTreeBuilder builder;
            builder.setColumns(int(Columns::KEY), int(Columns::VALUE), int(Columns::TYPE));
            builder.updateTree(treeItemData, *aronType, *data);
        }
        else
        {
            DataTreeBuilder builder;
            builder.setColumns(int(Columns::KEY), int(Columns::VALUE), int(Columns::TYPE));
            builder.updateTree(treeItemData, *data);
        }
        treeItemData->setExpanded(true);
    }

    void InstanceView::updateMetaData(const wm::EntityInstanceMetadata& metadata)
    {
        std::vector<std::string> items =
        {
            std::to_string(metadata.confidence),
            armem::toDateTimeMilliSeconds(metadata.timeCreated),
            armem::toDateTimeMilliSeconds(metadata.timeSent),
            armem::toDateTimeMilliSeconds(metadata.timeArrived)
        };
        ARMARX_CHECK_EQUAL(treeItemMetadata->childCount(), items.size());
        int i = 0;
        for (const std::string& item : items)
        {
            treeItemMetadata->child(i++)->setText(int(Columns::VALUE), QString::fromStdString(item));
        }
    }

    void InstanceView::showErrorMessage(const std::string& message)
    {
        if (statusLabel)
        {
            statusLabel->setText(QString::fromStdString(message));
        }
    }


    void InstanceView::prepareTreeContextMenu(const QPoint& pos)
    {
        QTreeWidgetItem* item = tree->itemAt(pos);
        if (item == nullptr || item->parent() == nullptr)
        {
            // No item or top level item => no context menu.
            return;
        }

        QMenu menu(this);

        aron::type::Descriptor type = static_cast<aron::type::Descriptor>(item->data(int(Columns::TYPE), Qt::UserRole).toInt());
        switch (type)
        {
            case aron::type::Descriptor::eIVTCByteImage:
            {
                QStringList qpath = item->data(int(Columns::KEY), Qt::UserRole).toStringList();
                aron::Path path = deserializePath(qpath);

                QAction* viewAction = new QAction("Show image");
                menu.addAction(viewAction);
                connect(viewAction, &QAction::triggered, [this, path]()
                {
                    this->showImageView(path);
                });
            }
            break;
            default:
                return;
        }

        menu.exec(tree->mapToGlobal(pos));
    }

    void InstanceView::showImageView(const aron::Path& elementPath)
    {
        if (!currentInstance)
        {
            return;
        }
        if (!imageView)
        {
            imageView = ImageView(splitter);
        }
        imageView->elementPath = elementPath;
        updateImageView(currentInstance->data());
    }

    void InstanceView::removeImageView()
    {
        imageView->group->hide();
        imageView->group->deleteLater();
        imageView = std::nullopt;
    }

    void InstanceView::updateImageView(const aron::datanavigator::DictNavigatorPtr& data)
    {
        using aron::datanavigator::NDArrayNavigator;

        if (!imageView)
        {
            return;
        }
        if (!data)
        {
            removeImageView();
            return;
        }

        aron::datanavigator::NavigatorPtr element;
        try
        {
            element = data->navigateAbsolute(imageView->elementPath);
        }
        // This can happen when the underlying entity structure changes (a new entity has been selected).
        // In this case, we disable the image view.
        catch (const aron::error::AronException&)
        {
            // showErrorMessage(e.what());
            removeImageView();
            return;
        }
        catch (const armarx::LocalException&)
        {
            // showErrorMessage(e.what());
            removeImageView();
            return;
        }

        NDArrayNavigator::PointerType imageData = NDArrayNavigator::DynamicCast(element);
        if (!imageData)
        {
            showErrorMessage("Expected NDArrayNavigator, but got: " + simox::meta::get_type_name(element));
            return;
        }

        std::vector<int> shape = imageData->getDimensions();
        if (shape.size() != 3)
        {
            showErrorMessage("Expected array shape with 3 dimensions, but got: "
                             + NDArrayNavigator::DimensionsToString(shape));
            return;
        }

        QImage::Format format = QImage::Format::Format_Invalid;
        switch (shape.at(2))
        {
            case 1:
                format = QImage::Format::Format_Grayscale8;
                break;
            case 3:
                format = QImage::Format::Format_RGB888;
                break;
            default:
                showErrorMessage("Expected 1 or 3 elements in last dimension, but got shape: "
                                 + NDArrayNavigator::DimensionsToString(shape));
                return;
        }

        std::stringstream title;
        title << "Image element '" << imageView->elementPath.toString() << "'"; // of entity instance " << currentInstance->id();
        imageView->group->setTitle(QString::fromStdString(title.str()));
        imageView->view->setImage(QImage(imageData->getData(), shape.at(0), shape.at(1), format));
    }


    InstanceView::ImageView::ImageView(QSplitter* parent)
    {
        group = new QGroupBox();
        parent->addWidget(group);

        group->setLayout(new QHBoxLayout());
        int margin = 2;
        group->layout()->setContentsMargins(margin, margin, margin, margin);
        if (false)
        {
            QFont font = group->font();
            font.setPointSizeF(font.pointSize() * 0.75);
            group->setFont(font);
        }

        view = new instance::ImageView();
        group->layout()->addWidget(view);
    }

}
