#include "Writer.h"

#include <mutex>
#include <optional>

#include "ArmarXCore/core/logging/Logging.h"

#include "RobotAPI/libraries/armem/core/MemoryID.h"
#include "RobotAPI/libraries/armem_objects/aron_conversions.h"
#include <RobotAPI/libraries/armem_objects/aron/RobotDescription.aron.generated.h>
#include <RobotAPI/libraries/armem_objects/aron/Robot.aron.generated.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>


namespace armarx::armem::articulated_object
{
    Writer::Writer(ManagedIceObject& component): MemoryConnector(component) {}

    void Writer::registerPropertyDefinitions(armarx::PropertyDefinitionsPtr& def)
    {
        ARMARX_DEBUG << "Writer: registerPropertyDefinitions";

        MemoryConnector::registerPropertyDefinitions(def);

        const std::string prefix = getPropertyPrefix();

        def->optional(properties.memoryName, prefix + "MemoryName");

        def->optional(properties.coreInstanceSegmentName,
                      prefix + "CoreSegment",
                      "Name of the memory core segment to use for object instances.");
        def->optional(properties.coreClassSegmentName,
                      prefix + "CoreSegment",
                      "Name of the memory core segment to use for object classes.");
        def->required(properties.providerName, prefix + "ProviderName");
    }

    void Writer::connect()
    {
        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "Writer: Waiting for memory '" << properties.memoryName << "' ...";
        auto result = useMemory(properties.memoryName);
        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;
            return;
        }

        ARMARX_IMPORTANT << "Writer: Connected to memory '" << properties.memoryName;

        memoryWriter.setWritingMemory(result.proxy);
        memoryReader.setReadingMemory(result.proxy);

        const auto resultCoreClassSegment = memoryWriter.addSegment(properties.coreClassSegmentName, properties.providerName);

        const auto resultCoreInstanceSegmentName =
            memoryWriter.addSegment(properties.coreInstanceSegmentName, properties.providerName);

        armem::MemoryID refId = armem::MemoryID(resultCoreClassSegment.segmentID);

        armem::MemoryID id;
        id.setCoreSegmentID(refId); // listen to all provider segments!

        memoryReader.subscribe(id, this, &Writer::updateKnownObjects);
    }

    void Writer::updateKnownObject(const armem::MemoryID& snapshotId)
    {

        arondto::RobotDescription aronArticulatedObjectDescription;
        // aronArticulatedObjectDescription.fromAron(snapshotId.ent);

        // TODO(fabian.reister): implement
    }

    void Writer::updateKnownObjects(const armem::MemoryID& subscriptionID, const std::vector<armem::MemoryID>& snapshotIDs)
    {
        ARMARX_INFO << "New objects available!";

        std::for_each(snapshotIDs.begin(), snapshotIDs.end(), [&](const auto & snapshotID)
        {
            updateKnownObject(snapshotID);
        });
    }

    std::optional<armem::MemoryID> Writer::storeOrGetClass(const ArticulatedObject& obj)
    {
        const auto objectId = knownObjects.find(obj.description.name);

        // check if exists
        if (objectId != knownObjects.end())
        {
            return objectId->second;
        }

        // otherwise create
        return storeClass(obj);
    }

    std::optional<armem::MemoryID> Writer::storeClass(const ArticulatedObject& obj)
    {
        std::lock_guard g{memoryWriterMutex};

        ARMARX_DEBUG << "Trying to create core segment + provider segment";

        // TODO(fabian.reister): variable provider segment
        const auto result = memoryWriter.addSegment(properties.coreClassSegmentName, properties.providerName);

        if (not result.success)
        {
            ARMARX_ERROR << "Creating core segment failed. Reason: " << result.errorMessage;
            return std::nullopt;
        }

        const auto& timestamp = obj.timestamp;

        const auto providerId = armem::MemoryID(result.segmentID);
        const auto entityID =
            providerId
            .withEntityName(obj.description.name)
            .withTimestamp(timestamp);

        armem::EntityUpdate update;
        update.entityID = entityID;

        arondto::RobotDescription aronArticulatedObjectDescription;
        toAron(aronArticulatedObjectDescription, obj.description);

        update.instancesData = {aronArticulatedObjectDescription.toAron()};
        update.timeCreated   = timestamp;

        ARMARX_DEBUG << "Committing " << update << " at time " << timestamp;
        armem::EntityUpdateResult updateResult = memoryWriter.commit(update);

        ARMARX_DEBUG << updateResult;

        if (not updateResult.success)
        {
            ARMARX_ERROR << updateResult.errorMessage;
            return std::nullopt;
        }

        // update cache (TODO: likely remove this)
        knownObjects[obj.description.name] = updateResult.snapshotID;

        return updateResult.snapshotID;
    }

    bool Writer::storeInstance(const ArticulatedObject& obj)
    {
        std::lock_guard g{memoryWriterMutex};

        ARMARX_DEBUG << "Trying to create core segment + provider segment";

        const auto result =
            memoryWriter.addSegment(properties.coreInstanceSegmentName, properties.providerName);

        if (not result.success)
        {
            ARMARX_ERROR << "Creating core segment failed. Reason: " << result.errorMessage;
            return false;
        }

        const auto& timestamp = obj.timestamp;

        const auto providerId = armem::MemoryID(result.segmentID);
        const auto entityID =
            providerId
            .withEntityName(obj.description.name)
            .withTimestamp(timestamp);

        armem::EntityUpdate update;
        update.entityID = entityID;

        arondto::Robot aronArticulatedObject;
        toAron(aronArticulatedObject, obj);

        const auto descriptionId = storeOrGetClass(obj);

        if (not descriptionId)
        {
            ARMARX_ERROR << "Could not get class for object " << obj.description.name;
            return false;
        }

        // install memory link
        toAron(aronArticulatedObject.description, *descriptionId);

        update.instancesData = {aronArticulatedObject.toAron()};
        update.timeCreated   = timestamp;

        ARMARX_DEBUG << "Committing " << update << " at time " << timestamp;
        armem::EntityUpdateResult updateResult = memoryWriter.commit(update);

        ARMARX_DEBUG << updateResult;

        if (not updateResult.success)
        {
            ARMARX_ERROR << updateResult.errorMessage;
        }

        return updateResult.success;
    }

    bool Writer::store(const ArticulatedObject& obj)
    {
        const std::optional<armem::MemoryID> classId = storeOrGetClass(obj);

        if (not classId)
        {
            ARMARX_WARNING << "Could not get class id!";
            return false;
        }

        // TODO(fabian.reister): integrate memory link
        return storeInstance(obj);
    }

    const std::string& Writer::getPropertyPrefix() const
    {
        return propertyPrefix;
    }

} // namespace armarx::armem::articulated_object