/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Pair.h"


namespace armarx::aron::typenavigator
{
    // constructors
    PairNavigator::PairNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::ePair, path),
        Navigator(type::Descriptor::ePair, path),
        type(new type::AronPair())
    {

    }

    PairNavigator::PairNavigator(const type::AronPairPtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::ePair, path),
        Navigator(type::Descriptor::ePair, path),
        type(o)
    {
        CheckAronPtrForNull("AronPairTypeNavigator", "AronPairTypeNavigator", getPath(), o);

        acceptedType1 = FromAronType(o->acceptedType1);
        acceptedType2 = FromAronType(o->acceptedType2);
    }

    // Member functions
    std::pair<NavigatorPtr, NavigatorPtr> PairNavigator::getAcceptedTypes() const
    {
        return {acceptedType1, acceptedType2};
    }

    NavigatorPtr PairNavigator::getFirstAcceptedType() const
    {
        return acceptedType1;
    }

    NavigatorPtr PairNavigator::getSecondAcceptedType() const
    {
        return acceptedType2;
    }

    void PairNavigator::addAcceptedType(const NavigatorPtr& n)
    {
        if (acceptedType1 == nullptr)
        {
            acceptedType1 = n;
            return;
        }
        if (acceptedType2 == nullptr)
        {
            acceptedType2 = n;
            return;
        }
        throw error::AronException("PairNavigator", "addAcceptedType", "Both types are already set", getPath());
    }

    void PairNavigator::setFirstAcceptedType(const NavigatorPtr& n)
    {
        acceptedType1 = n;
    }

    void PairNavigator::setSecondAcceptedType(const NavigatorPtr& n)
    {
        acceptedType2 = n;
    }

    type::AronPairPtr PairNavigator::toAronPairPtr() const
    {
        CheckAronPtrForNull("PairNavigator", "getResult", getPath(), type->acceptedType1);
        CheckAronPtrForNull("PairNavigator", "getResult", getPath(), type->acceptedType2);
        return type;
    }

    // static methods
    PairNavigatorPtr PairNavigator::DynamicCast(const NavigatorPtr& n)
    {
        return std::dynamic_pointer_cast<PairNavigator>(n);
    }

    PairNavigatorPtr PairNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckTypeNavigatorPtrForNull("AronPairTypeNavigator", "DynamicCastAndCheck[Before]", n);
        PairNavigatorPtr casted = PairNavigator::DynamicCast(n);
        CheckTypeNavigatorPtrForNull("AronPairTypeNavigator", "DynamicCastAndCheck[After]", n->getPath(), casted);
        return casted;
    }

    // virtual implementations
    std::vector<NavigatorPtr> PairNavigator::getChildren() const
    {
        return {acceptedType1, acceptedType2};
    }

    size_t PairNavigator::childrenSize() const
    {
        return 2;
    }

    type::AronTypePtr PairNavigator::getResult() const
    {
        CheckAronPtrForNull("PairNavigator", "getResult", getPath(), type->acceptedType1);
        CheckAronPtrForNull("PairNavigator", "getResult", getPath(), type->acceptedType2);
        return type;
    }

    std::string PairNavigator::getName() const
    {
        return "AronPairType<" + acceptedType1->getName() + ", " + acceptedType2->getName() + ">";
    }
}

