/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "OpenCVMat.h"

namespace armarx::aron::typenavigator
{
    // constructors
    OpenCVMatNavigator::OpenCVMatNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eOpenCVMat, path),
        Navigator(type::Descriptor::eOpenCVMat, path),
        type(new type::AronOpenCVMat({}, ""))
    {
    }

    OpenCVMatNavigator::OpenCVMatNavigator(const type::AronOpenCVMatPtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eOpenCVMat, path),
        Navigator(type::Descriptor::eOpenCVMat, path),
        type(o)
    {
        CheckAronPtrForNull("AronOpenCVMatTypeNavigator", "AronOpenCVMatTypeNavigator", getPath(), o);
        checkDimensions(type->dimensions);
        checkTypename(type->typeName);
    }

    bool OpenCVMatNavigator::checkDimensions(const std::vector<int>& d) const
    {
        if (d.size() < ACCEPTED_DIMENSION_MIN_SIZE)
        {
            throw error::AronException("AronOpenCVMatTypeNavigator", "checkDimensions", "The dimension size is wrong. Got size: " + std::to_string(d.size()), getPath());
        }

        if (std::any_of(d.begin(), d.end(), [](int i)
    {
        return i < -1 || i == 0;
    }))
        {
            throw error::AronException("AronOpenCVMatTypeNavigator", "checkDimensions", "The dimension size is wrong. At least one empty is < -1 or 0. Got as dimension: " + simox::alg::to_string(d, ", "), getPath());
        }
        return true;
    }

    std::string OpenCVMatNavigator::checkTypename(const std::string& s) const
    {
        for (const auto& [key, list] : ACCEPTED_TYPES)
        {
            if (s == key)
            {
                return key;
            }

            for (const auto& el : list)
            {
                if (s == el)
                {
                    return key;
                }
            }
        }
        throw error::AronException("OpenCVMatNavigator", "checkTypename", "The typename is empty.", getPath());
    }

    type::AronOpenCVMatPtr OpenCVMatNavigator::toAronOpenCVMatPtr() const
    {
        checkDimensions(type->dimensions);
        checkTypename(type->typeName);
        return type;
    }

    std::string OpenCVMatNavigator::getTypename() const
    {
        return type->typeName;
    }

    std::vector<int> OpenCVMatNavigator::getDimensions() const
    {
        return type->dimensions;
    }

    void OpenCVMatNavigator::setTypename(const std::string& u)
    {
        type->typeName = checkTypename(u);
    }

    void OpenCVMatNavigator::setDimensions(const std::vector<int>& d)
    {
        checkDimensions(d);
        type->dimensions = d;
    }


    // static methods
    OpenCVMatNavigatorPtr OpenCVMatNavigator::DynamicCast(const NavigatorPtr& n)
    {
        return std::dynamic_pointer_cast<OpenCVMatNavigator>(n);
    }

    OpenCVMatNavigatorPtr OpenCVMatNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckTypeNavigatorPtrForNull("AronOpenCVMatTypeNavigator", "DynamicCast[Before]", n);
        OpenCVMatNavigatorPtr casted = OpenCVMatNavigator::DynamicCast(n);
        CheckTypeNavigatorPtrForNull("AronOpenCVMatTypeNavigator", "DynamicCast[After]", n->getPath(), casted);
        return casted;
    }

    // virtual implementations
    std::vector<NavigatorPtr> OpenCVMatNavigator::getChildren() const
    {
        return {};
    }

    size_t OpenCVMatNavigator::childrenSize() const
    {
        return 0;
    }

    type::AronTypePtr OpenCVMatNavigator::getResult() const
    {
        return toAronOpenCVMatPtr();
    }

    std::string OpenCVMatNavigator::getName() const
    {
        return "AronOpenCVMatType<" + simox::alg::to_string(type->dimensions, ", ") + ", " + type->typeName + ">";
    }
}

